import re

from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarSectionLowercase(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.notneededspace",
                         severity="info",
                         message="Space at the beginning of the var is not needed")

    def __getMatches(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER)
        for i in [x for x in items if "append" not in x.SubItems]:
            if i.VarValueStripped.startswith(" "):
                res.append(i)
        return res

    def check(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i.Origin, i.InFileLine)
        return res

    def fix(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            i.Raw = re.sub(r"\"\s+", "\"", i.Raw) + "\n"
            i.VarValue = re.sub(r"\"\s+", "\"", i.VarValue) + "\n"
            res.append(_file)
        return res



from oelint_adv.cls_item import MissingFile
from oelint_adv.cls_rule import Rule


class FileIncludeNotFound(Rule):
    def __init__(self):
        super().__init__(id="oelint.file.includenotfound",
                         severity="warning",
                         message="'{FILE}' was not found")

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file,
                                      classifier=MissingFile.CLASSIFIER):
            if item.Statement == "include":
                res += self.finding(item.Origin, item.InFileLine, self.Msg.replace("{FILE}", item.Filename))
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.const_vars import VAR_ORDER


class VarsOrder(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.order",
                         severity="warning",
                         message="<FOO>")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for item in items:
            _func_before = sorted(
                [x for x in items if x.Line < item.Line and x.VarName in VAR_ORDER], key=lambda x: x.Line, reverse=True)
            if any(_func_before):
                _func_before = _func_before[0]
                if item.VarName not in VAR_ORDER:
                    continue
                if VAR_ORDER.index(item.VarName) < VAR_ORDER.index(_func_before.VarName):
                    res += self.finding(item.Origin, item.InFileLine,
                                        "'{}' should be placed before '{}'".format(
                                            item.VarName, _func_before.VarName))
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.helper_files import safe_linesplit


class VarDuplicates(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.duplicate",
                         severity="warning",
                         message="<FOO>")

    def check(self, _file, stash):
        res = []
        for c in ["DEPENDS", "RDEPENDS_${PN}"]:
            items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                      attribute=Variable.ATTR_VAR, attributeValue=c)
            _items = []
            for i in items:
                for x in [y for y in safe_linesplit(i.VarValueStripped) if y]:
                    if x in _items:
                        res += self.finding(i.Origin, i.InFileLine,
                                            "Item '{}' was added multiple timesto {}".format(x, c))
                    else:
                        _items.append(x)
        return res



import re

from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.helper_files import safe_linesplit


class VarImproperInherit(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.improperinherit",
                         severity="error",
                         message="'{INH}' is not a proper bbclass name")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="inherit")
        for i in items:
            for subi in [x for x in safe_linesplit(i.VarValueStripped) if x]:
                if not re.match(r"^[A-Za-z0-9_.-]+$", subi):
                    res += self.finding(i.Origin, i.InFileLine,
                                        self.Msg.replace("{INH}", subi))
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.helper_files import get_scr_components, safe_linesplit


class VarSRCUriOptions(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.srcuridomains",
                         severity="warning",
                         message="Recipe is pulling from different domains, this will likely cause issues")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")

        _domains = set()
        for i in items:
            for u in [x.strip() for x in safe_linesplit(i.VarValueStripped)]:
                _url = get_scr_components(u)
                if _url["scheme"] and _url["scheme"] not in ["file"]:
                    _domains.add(_url["src"].split("/")[0])
        if len(_domains) > 1:
            res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.const_vars import get_mandatory_vars


class VarMandatoryExists(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.mandatoryvar",
                         severity="error",
                         message="<FOO>",
                         onappend=False)

    def check(self, _file, stash):
        res = []
        for var in get_mandatory_vars():
            items = stash.GetItemsFor(
                filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue=var)
            if not any(items):
                res += self.finding(_file, 0, "Variable '{}' should be set".format(var))
        return res



_RULE_FILE = {}
_CONST_FILE = {}


def get_rulefile():
    return _RULE_FILE


def get_constantfile():
    return _CONST_FILE


def set_rulefile(value):
    global _RULE_FILE
    _RULE_FILE = value


def set_constantfile(value):
    global _CONST_FILE
    _CONST_FILE = value



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.const_vars import get_protected_vars


class VarQuoted(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.bbvars",
                         severity="warning",
                         message="Variable '{VAR}' should be set on a disto/layer or local.conf level, not in a recipe")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in [x for x in items if x.VarName in get_protected_vars()]:
            if i.VarOp not in [" ??= ", " ?= "]:
                res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.replace("{VAR}", i.VarName))
        return res



import textwrap
import re

from oelint_adv.const_func import KNOWN_FUNCS


class Item():
    ATTR_LINE = "Line"
    ATTR_RAW = "Raw"
    ATTR_ORIGIN = "Origin"
    CLASSIFIER = "Item"
    ATTR_SUB = "SubItem"

    def __init__(self, origin, line, infileline, rawtext):
        self.Line = line
        self.Raw = rawtext
        self.Links = []
        self.Origin = origin
        self.InFileLine = infileline

    def extract_sub(self, name):
        chunks = name.split("_")
        _suffix = []
        _var = []
        for i in chunks:
            if "-" in i:
                # just use the prefix in case a dash is found
                # that addresses things like FILES_${PN}-dev
                i = i.split("-")[0]
            if re.match("^[A-Z0-9{}$]+$", i):
                _var.append(i)
            else:
                _suffix.append(i)
        _var = [x for x in _var if x]
        _suffix = [x for x in _suffix if x]
        if not _var and _suffix:
            # special case for pkg-functions
            _var = _suffix
            _suffix = []
        return ("_".join(_var), "_".join(_suffix))

    def extract_sub_func(self, name):
        chunks = name.split("_")
        _marker = ["append", "prepend", "class-native",
                   "class-cross", "class-target", "remove"]
        _suffix = []
        _var = []
        for i in chunks:
            if i in _marker or "_".join(_var) in KNOWN_FUNCS:
                _suffix = chunks[chunks.index(i):]
                break
            else:
                _var.append(i)
        _var = [x for x in _var if x]
        _suffix = [x for x in _suffix if x]
        return ("_".join(_var), "_".join(_suffix))

    def IsFromAppend(self):
        return self.Origin.endswith(".bbappend")

    def AddLink(self, _file):
        self.Links.append(_file)
        self.Links = list(set(self.Links))

    def GetAttributes(self):
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}

    def GetRawCleaned(self, chars="\t\n\r"):
        res = self.Raw
        for c in chars:
            res = res.replace(c, " ")
        return res

    def __repr__(self):
        return "{} -- {}\n".format(self.__class__.__name__, self.GetAttributes())


class Variable(Item):
    ATTR_VAR = "VarName"
    ATTR_VARVAL = "VarValue"
    ATTR_VARVALSTRIPPED = "VarValueStripped"
    CLASSIFIER = "Variable"
    VAR_VALID_OPERATOR = [" = ", " += ",
                          " ?= ", " ??= ", " := ", " .= ", " =+ "]

    def __init__(self, origin, line, infileline, rawtext, name, value, operator, flag):
        super().__init__(origin, line, infileline, rawtext)
        if "inherit" != name:
            self.VarName, self.SubItem = self.extract_sub(name)
        else:
            self.VarName = name
            self.SubItem = ""
        self.SubItems = self.SubItem.split("_")
        self.VarValue = value
        self.VarOp = operator
        self.Flag = flag
        self.VarValueStripped = self.VarValue.strip().lstrip('"').rstrip('"')

    def IsAppend(self):
        return self.VarOp in [" += "] or "append" in self.SubItems

    def AppendOperation(self):
        res = []
        if self.VarOp in [" += "]:
            res.append(self.VarOp)
        if "append" in self.SubItems:
            res.append("append")
        if "prepend" in self.SubItems:
            res.append("prepend")
        if "remove" in self.SubItems:
            res.append("remove")
        return res

    def IsMultiLine(self):
        return "\\" in self.Raw

    def GetMachineEntry(self):
        for x in self.SubItems:
            if x not in ["append", "prepend", "class-native", "class-cross", "class-target", "remove", "machine"]:
                return x
        return ""


class Comment(Item):
    CLASSIFIER = "Comment"

    def __init__(self, origin, line, infileline, rawtext):
        super().__init__(origin, line, infileline, rawtext)


class Include(Item):
    CLASSIFIER = "Include"
    ATTR_INCNAME = "IncName"
    ATTR_STATEMENT = "Statement"

    def __init__(self, origin, line, infileline, rawtext, incname, statement):
        super().__init__(origin, line, infileline, rawtext)
        self.IncName = incname
        self.Statement = statement


class Function(Item):
    ATTR_FUNCNAME = "FuncName"
    ATTR_FUNCBODY = "FuncBody"
    CLASSIFIER = "Function"

    def __init__(self, origin, line, infileline, rawtext, name, body, python=False, fakeroot=False):
        super().__init__(origin, line, infileline, rawtext)
        self.IsPython = python is not None
        self.IsFakeroot = fakeroot is not None
        name = name or ""
        self.FuncName, self.SubItem = self.extract_sub_func(name.strip())
        self.SubItems = self.SubItem.split("_")
        self.FuncBody = body
        self.FuncBodyStripped = body.replace(
            "{", "").replace("}", "").replace("\n", "").strip()
        self.FuncBodyRaw = textwrap.dedent(
            rawtext[rawtext.find("{") + 1:].rstrip().rstrip("}"))

    def GetMachineEntry(self):
        for x in self.SubItems:
            if x not in ["append", "prepend", "class-native", "class-cross", "class-target", "remove", "machine"]:
                return x
        return ""


class PythonBlock(Item):
    ATTR_FUNCNAME = "FuncName"
    CLASSIFIER = "PythonBlock"

    def __init__(self, origin, line, infileline, rawtext, name):
        super().__init__(origin, line, infileline, rawtext)
        self.FuncName = name


class TaskAssignment(Item):
    ATTR_FUNCNAME = "FuncName"
    ATTR_VAR = "VarName"
    ATTR_VARVAL = "VarValue"
    CLASSIFIER = "TaskAssignment"

    def __init__(self, origin, line, infileline, rawtext, name, ident, value):
        super().__init__(origin, line, infileline, rawtext)
        self.FuncName = name
        self.VarName = ident
        self.VarValue = value


class TaskAdd(Item):
    ATTR_FUNCNAME = "FuncName"
    ATTR_BEFORE = "Before"
    ATTR_AFTER = "After"
    CLASSIFIER = "TaskAdd"

    def __init__(self, origin, line, infileline, rawtext, name, before="", after=""):
        super().__init__(origin, line, infileline, rawtext)
        self.FuncName = name
        self.Before = [x for x in (before or "").split(" ") if x]
        self.After = [x for x in (after or "").split(" ") if x]


class MissingFile(Item):
    ATTR_FILENAME = "Filename"
    ATTR_STATEMENT = "Statement"
    CLASSIFIER = "MissingFile"

    def __init__(self, origin, line, infileline, filename, statement):
        super().__init__(origin, line, infileline, "")
        self.Filename = filename
        self.Statement = statement



import importlib
import inspect
import os
import pkgutil

from colorama import Fore, Style

from oelint_adv.color import get_color
from oelint_adv.rule_file import get_rulefile


class Rule():
    def __init__(self, id="", severity="", message="", onappend=True, onlyappend=False):
        self.ID = id
        self.Severity = severity
        self.Msg = message
        self.OnAppend = onappend
        self.OnlyAppend = onlyappend

    def check(self, _file, stash):
        return []

    def fix(self, _file, stash):
        return []

    def finding(self, _file, _line, override_msg=None):
        if not self.OnAppend and _file.endswith(".bbappend"):
            return
        if self.OnlyAppend and not _file.endswith(".bbappend"):
            return
        if override_msg is None:
            override_msg = self.Msg
        _severity = self.Severity
        _rule_file = get_rulefile()
        if _rule_file and self.ID in _rule_file:
            _severity = _rule_file[self.ID] or self.Severity
        if _line <= 0:
            # Fix those issues, that don't come with a line
            _line = 1
        if get_color():
            if _severity == "error":
                return ["{}:{}{}:{}:{}:{}{}".format(os.path.abspath(_file), Fore.RED, _line, _severity, self.ID, override_msg, Style.RESET_ALL)]
            elif _severity == "warning":
                return ["{}:{}{}:{}:{}:{}{}".format(os.path.abspath(_file), Fore.YELLOW, _line, _severity, self.ID, override_msg, Style.RESET_ALL)]
            else:
                return ["{}:{}{}:{}:{}:{}{}".format(os.path.abspath(_file), Fore.GREEN, _line, _severity, self.ID, override_msg, Style.RESET_ALL)]
        return ["{}:{}:{}:{}:{}".format(os.path.abspath(_file), _line, _severity, self.ID, override_msg)]

    def __repr__(self):
        return "{}".format(self.ID)

    def FormatMsg(self, *args):
        return self.Msg.format(*args)


def load_rules(add_rules=[]):
    res = []
    _rule_file = get_rulefile()
    _path_list = {
        "base": {"path": "rule_base"}
    }
    for ar in add_rules:
        _path_list[ar] = {"path": "rule_{}".format(ar)}
    for _, v in _path_list.items():
        _searchpath = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), v["path"])
        packages = pkgutil.walk_packages(path=[_searchpath])
        for _, name, _ in packages:
            name = __name__.split(".")[0] + "." + v["path"] + "." + name
            mod = importlib.import_module(name)
            for m in inspect.getmembers(mod, inspect.isclass):
                try:
                    if issubclass(m[1], Rule):
                        inst = m[1]()
                        if inst.ID:
                            if _rule_file and inst.ID not in _rule_file:
                                continue
                            res.append(inst)
                except Exception:
                    pass
    return res



from copy import deepcopy

from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarDependsOrdered(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.dependsordered",
                         severity="warning",
                         message="[R]DEPENDS entries should be ordered alphabetically")

    def __get_tuple_wildcard_index(self, _list, elem):
        for i in range(len(_list)):
            if _list[i][1] == elem:
                return i
        return -1

    def check(self, _file, stash):
        res = []
        for elem in ["DEPENDS", "RDEPENDS_${PN}"]:
            items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                      attribute=Variable.ATTR_VAR, attributeValue=elem)
            items = sorted(items, key=lambda x: x.Line)
            _findings = []
            for i in items:
                linenum = 0
                for line in i.VarValueStripped.replace(chr(0x1b), "").split("\n"):
                    for x in [x for x in line.split(" ") if x]:
                        _i = deepcopy(i)
                        _i.Line += linenum
                        _findings.append((_i, x))
                    linenum += 1
            _sorted = sorted(_findings, key=lambda tup: tup[1])
            for i in _sorted:
                if self.__get_tuple_wildcard_index(_sorted, i[1]) != \
                        self.__get_tuple_wildcard_index(_findings, i[1]):
                    res += self.finding(i[0].Origin, i[0].InFileLine)
        return res



import glob
import os
import re
from urllib.parse import urlparse

from oelint_adv.cls_item import Variable
from oelint_adv.const_vars import get_known_mirrors


def get_files(stash, _file, pattern):
    res = []
    src_uris = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                 attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")
    files_paths = list(
        set(["{}/*/{}".format(os.path.dirname(x.Origin), pattern) for x in src_uris]))
    for item in src_uris:
        files_paths += list(set(["{}/*/{}".format(os.path.dirname(x.Origin), pattern)
                                 for x in stash.GetItemsFor(filename=item.Origin)]))
    for item in files_paths:
        res += glob.glob(item)
    return list(set(res))


def find_local_or_in_layer(name, localdir):
    if os.path.exists(os.path.join(localdir, name)):
        return os.path.join(localdir, name)
    _curdir = localdir
    while os.path.isdir(_curdir):
        if _curdir == "/":
            break
        _curdir = os.path.dirname(_curdir)
        if os.path.exists(os.path.join(_curdir, "conf/layer.conf")):
            if os.path.exists(os.path.join(_curdir, name)):
                return os.path.join(_curdir, name)
            else:
                break
    return None


def _replace_with_known_mirrors(_in):
    """
    Replace the known mirror configuration items
    """
    for k, v in get_known_mirrors().items():
        _in = _in.replace(k, v)
    return _in


def get_scr_components(string):
    """
        Parses an URL string
        returns a dict with
            scheme = protcol used
            src = path to call
            options = dict with options added to URL
    """
    _url = urlparse(_replace_with_known_mirrors(string))
    _scheme = _url.scheme
    _options = _url.netloc.split(";")[1:]
    _path = _url.netloc.split(";")[0]
    if _url.path:
        if not _path.endswith("/") and not _url.path.startswith("/"):
            _path += "/"
        _path += _url.path
    _parsed_opt = {x.split("=")[0]: x.split("=")[1] for x in _options}
    return {"scheme": _scheme, "src": _path, "options": _parsed_opt}


def safe_linesplit(string):
    return re.split(r"\s|\t|\x1b", string)



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarHomepagePrefix(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.homepageprefix",
                         severity="warning",
                         message="'HOMEPAGE' should start with 'http://' or 'https://'")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="HOMEPAGE")
        for i in items:
            if not any([x for x in ["https://", "http://"] if i.VarValueStripped.startswith(x)]):
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.rule_file import get_rulefile, get_constantfile


def __get_const_and_rulefile(l, key):
    res = l
    _rulefile = {**get_rulefile(), **get_constantfile()}
    if key in _rulefile.keys() and isinstance(_rulefile[key], list):
        res += _rulefile[key]
    return res


def __get_const_and_rulefile_dict(l, key):
    res = l
    _rulefile = {**get_rulefile(), **get_constantfile()}
    if key in _rulefile.keys() and isinstance(_rulefile[key], dict):
        res = {**res, **_rulefile[key]}
    return res


MANDATORY_VARS = [
    "SUMMARY",
    "DESCRIPTION",
    "AUTHOR",
    "HOMEPAGE",
    "SECTION",
    "LICENSE",
    "SRC_URI"
]


def get_mandatory_vars():
    return __get_const_and_rulefile(MANDATORY_VARS, "mandatory_vars")


SUGGESTED_VARS = [
    "BUGTRACKER",
    "BBCLASSEXTEND",
    "CVE_PRODUCT"
]


def get_suggested_vars():
    return __get_const_and_rulefile(SUGGESTED_VARS, "suggested_vars")


KNOWN_MIRRORS = {
    "${APACHE_MIRROR}": "http://www.us.apache.org/dist",
    "${CPAN_MIRROR}": "http://cpan.metacpan.org/",
    "${DEBIAN_MIRROR}": "http://snapshot.debian.org/archive/debian/20180310T215105Z/pool",
    "${E_CVS}": "cvs://anonymous@anoncvs.enlightenment.org/var/cvs/e",
    "${E_URI}": "http://enlightenment.freedesktop.org/files",
    "${FREEBSD_MIRROR}": "ftp://ftp.freebsd.org/pub/FreeBSD/",
    "${FREEDESKTOP_CVS}": "cvs://anoncvs:anoncvs@anoncvs.freedesktop.org/cvs",
    "${GENTOO_MIRROR}": "http://distro.ibiblio.org/pub/linux/distributions/gentoo/distfiles",
    "${GNOME_MIRROR}": "http://ftp.gnome.org/pub/GNOME/sources",
    "${GNU_MIRROR}": "https://mirrors.kernel.org/gnu",
    "${GNUPG_MIRROR}": "ftp://ftp.gnupg.org/gcrypt",
    "${GPE_MIRROR}": "http://gpe.linuxtogo.org/download/source",
    "${GPE_SVN}": "svn://projects.linuxtogo.org/svn/gpe/trunk/base;module=${PN}",
    "${HANDHELDS_CVS}": "cvs://anoncvs:anoncvs@anoncvs.handhelds.org/cvs",
    "${KERNELORG_MIRROR}": "http://www.kernel.org/pub",
    "${SAVANNAH_GNU_MIRROR}": "http://download-mirror.savannah.gnu.org/releases",
    "${SAVANNAH_NONGNU_MIRROR}": "http://download-mirror.savannah.nongnu.org/releases",
    "${SOURCEFORGE_MIRROR}": "http://heanet.dl.sourceforge.net/sourceforge",
    "${XLIBS_MIRROR}": "http://xlibs.freedesktop.org/release",
    "${XORG_MIRROR}": "http://xorg.freedesktop.org/releases"
}


def get_known_mirrors():
    return __get_const_and_rulefile_dict(KNOWN_MIRRORS, "known_mirrors")


VAR_PROTECTED = [
    "BB_CONSOLELOG",
    "BB_CURRENTTASK",
    "BB_DANGLINGAPPENDS_WARNONLY",
    "BB_DEFAULT_TASK",
    "BB_DISKMON_DIRS",
    "BB_DISKMON_WARNINTERVAL",
    "BB_ENV_EXTRAWHITE",
    "BB_ENV_WHITELIST",
    "BB_FETCH_PREMIRRORONLY",
    "BB_FILENAME",
    "BB_GENERATE_MIRROR_TARBALLS",
    "BB_HASHBASE_WHITELIST",
    "BB_HASHCHECK_FUNCTION",
    "BB_HASHCONFIG_WHITELIST",
    "BB_INVALIDCONF",
    "BB_LOGFMT",
    "BB_NICE_LEVEL",
    "BB_NO_NETWORK",
    "BB_NUMBER_PARSE_THREADS",
    "BB_NUMBER_THREADS",
    "BB_ORIGENV",
    "BB_PRESERVE_ENV",
    "BB_RUNFMT",
    "BB_RUNTASK",
    "BB_SCHEDULER",
    "BB_SCHEDULERS",
    "BB_SETSCENE_DEPVALID",
    "BB_SETSCENE_VERIFY_FUNCTION",
    "BB_SIGNATURE_EXCLUDE_FLAGS",
    "BB_SIGNATURE_HANDLER",
    "BB_SRCREV_POLICY",
    "BB_STAMP_POLICY",
    "BB_STAMP_WHITELIST",
    "BB_STRICT_CHECKSUM",
    "BB_TASK_NICE_LEVEL",
    "BB_TASKHASH",
    "BB_VERBOSE_LOGS",
    "BB_WORKERCONTEXT",
    "BBDEBUG",
    "BBFILE_COLLECTIONS",
    "BBFILE_PATTERN",
    "BBFILE_PRIORITY",
    "BBFILES",
    "BBINCLUDED",
    "BBINCLUDELOGS",
    "BBINCLUDELOGS_LINES",
    "BBLAYERS",
    "BBMASK",
    "BBPATH",
    "BBSERVER",
    "BBVERSIONS",
    "BITBAKE_UI",
    "BUILDNAME",
    "CACHE",
    "DEFAULT_PREFERENCE",
    "DL_DIR",
    "FILE",
    "FILESDIR",
    "FILESPATH",
    "INHERIT",
    "LAYERDEPENDS",
    "LAYERDIR",
    "LAYERVERSION",
    "MIRRORS",
    "MULTI_PROVIDER_WHITELIST",
    "PERSISTENT_DIR",
    "PREFERRED_PROVIDER",
    "PREFERRED_PROVIDERS",
    "PREFERRED_VERSION",
    "PREMIRRORS",
    "PRSERV_HOST",
    "STAMP",
    "TOPDIR"
]


def get_protected_vars():
    return __get_const_and_rulefile(VAR_PROTECTED, "protected_vars")


VAR_PROTECTED_APPEND = [
    "PV",
    "PR",
    "SRCREV",
    "LICENSE",
    "LIC_FILES_CHKSUM"
]


def get_protected_append_vars():
    return __get_const_and_rulefile(VAR_PROTECTED_APPEND, "protected_append_vars")


VAR_ORDER = [
    "SUMMARY",
    "DESCRIPTION",
    "AUTHOR",
    "HOMEPAGE",
    "BUGTRACKER",
    "SECTION",
    "LICENSE",
    "LIC_FILES_CHKSUM",
    "DEPENDS",
    "PROVIDES",
    "PV",
    "SRC_URI",
    "SRCREV",
    "S",
    "inherit",
    "PACKAGECONFIG",
    "EXTRA_QMAKEVARS_POST",
    "EXTRA_OECONF",
    "PACKAGE_ARCH",
    "PACKAGES",
    "FILES_${PN}",
    "RDEPENDS_${PN}",
    "RRECOMMENDS_${PN}",
    "RSUGGESTS_${PN}",
    "RPROVIDES_${PN}",
    "RCONFLICTS_${PN}",
    "BBCLASSEXTEND"
]

KNOWN_VARS = [
    "ABIEXTENSION",
    "ALL_MULTILIB_PACKAGE_ARCHS",
    "ALL_QA",
    "ALTERNATIVE_PRIORITY_${PN}",
    "ALTERNATIVE_PRIORITY",
    "ANY_OF_DISTRO_FEATURES",
    "APACHE_MIRROR",
    "APPEND",
    "ARCH_DEFAULT_KERNELIMAGETYPE",
    "ASNEEDED",
    "ASSUME_PROVIDED",
    "AUTO_LIBNAME_PKGS",
    "AUTOREV",
    "AVAILTUNES",
    "B",
    "BAD_RECOMMENDATIONS",
    "BASE_DEFAULT_DEPS",
    "BASE_GOARM",
    "BASE_WORKDIR",
    "BASEDEPENDS",
    "BASELIB",
    "BB_CMDLINE",
    "BB_CONSOLELOG",
    "BB_CURRENT_MC",
    "BB_CURRENTTASK",
    "BB_DANGLINGAPPENDS_WARNONLY",
    "BB_DEFAULT_TASK",
    "BB_DISKMON_DIRS",
    "BB_DISKMON_WARNINTERVAL",
    "BB_ENV_EXTRAWHITE",
    "BB_ENV_WHITELIST",
    "BB_FETCH_PREMIRRORONLY",
    "BB_FILENAME",
    "BB_GENERATE_MIRROR_TARBALLS",
    "BB_HASHBASE_WHITELIST",
    "BB_HASHCHECK_FUNCTION",
    "BB_HASHCONFIG_WHITELIST",
    "BB_HASHFILENAME",
    "BB_INVALIDCONF",
    "BB_LOGFMT",
    "BB_MIN_VERSION",
    "BB_NICE_LEVEL",
    "BB_NO_NETWORK",
    "BB_NUMBER_PARSE_THREADS",
    "BB_NUMBER_THREADS",
    "BB_ORIGENV",
    "BB_PRESERVE_ENV",
    "BB_RUNFMT",
    "BB_RUNTASK",
    "BB_SCHEDULER",
    "BB_SCHEDULERS",
    "BB_SETSCENE_DEPVALID",
    "BB_SETSCENE_VERIFY_FUNCTION",
    "BB_SIGNATURE_EXCLUDE_FLAGS",
    "BB_SIGNATURE_HANDLER",
    "BB_SRCREV_POLICY",
    "BB_STAMP_POLICY",
    "BB_STAMP_WHITELIST",
    "BB_STRICT_CHECKSUM",
    "BB_TASK_NICE_LEVEL",
    "BB_TASKHASH",
    "BB_UNIHASH",
    "BB_VERBOSE_LOGS",
    "BB_VERSION",
    "BB_WORKERCONTEXT",
    "BBDEBUG",
    "BBFILE_COLLECTIONS",
    "BBFILE_PATTERN",
    "BBFILE_PRIORITY",
    "BBFILES_DYNAMIC",
    "BBFILES",
    "BBINCLUDED",
    "BBINCLUDELOGS_LINES",
    "BBINCLUDELOGS",
    "BBLAYERS_CONF_UPDATE_FUNCS",
    "BBLAYERS_FETCH_DIR",
    "BBLAYERS_LAYERINDEX_URL",
    "BBLAYERS",
    "BBMASK",
    "BBPATH",
    "BBSERVER",
    "BBVERSIONS",
    "BINUVERSION",
    "BITBAKE_UI",
    "BP",
    "BPN",
    "BRANCH",
    "BUGTRACKER",
    "BUILD_ARCH",
    "BUILD_AS_ARCH",
    "BUILD_CC_ARCH",
    "BUILD_CFLAGS",
    "BUILD_CPPFLAGS",
    "BUILD_CXXFLAGS",
    "BUILD_EXEEXT",
    "BUILD_GOARCH",
    "BUILD_GOOS",
    "BUILD_GOTUPLE",
    "BUILD_LD_ARCH",
    "BUILD_LDFLAGS",
    "BUILD_OPTIMIZATION",
    "BUILD_OS",
    "BUILD_PREFIX",
    "BUILD_SYS",
    "BUILD_VENDOR",
    "BUILDCFG_FUNCS",
    "BUILDCFG_HEADER",
    "BUILDCFG_NEEDEDVARS",
    "BUILDCFG_VARS",
    "BUILDNAME",
    "BUILDSDK_CFLAGS",
    "BUILDSDK_CPPFLAGS",
    "BUILDSDK_CXXFLAGS",
    "BUILDSDK_LDFLAGS",
    "BUILDSTATS_BASE",
    "BUSYBOX_SPLIT_SUID",
    "CACHE",
    "CCACHE_DISABLE",
    "CCACHE",
    "CGO_CFLAGS",
    "CGO_CPPFLAGS",
    "CGO_CXXFLAGS",
    "CGO_LDFLAGS",
    "CHRPATH_BIN",
    "CLANGSDK",
    "CLASSOVERRIDE",
    "CLEAN_BROKEN",
    "CLEANBROKEN",
    "CLEANFUNCS",
    "CMDLINE",
    "COMBINED_FEATURES",
    "COMMERCIAL_AUDIO_PLUGINS",
    "COMMERCIAL_VIDEO_PLUGINS",
    "COMMON_LICENSE_DIR",
    "COMPATIBLE_MACHINE",
    "COMPILER_RT",
    "COMPONENTS_DIR",
    "COMPRESSIONTYPES",
    "CONF_VERSION",
    "CONFIG_EXTRA_CFLAGS",
    "CONFIG_FDISK_SUPPORT_LARGE_DISKS",
    "CONFIG_LFS",
    "CONFIGURESTAMPFILE",
    "CONLOG",
    "CONNECTIVITY_CHECK_URIS",
    "CONVERSIONTYPES",
    "CORE_IMAGE_BASE_INSTALL",
    "CORE_IMAGE_EXTRA_INSTALL",
    "COREBASE_FILES",
    "COREBASE",
    "CPAN_MIRROR",
    "CPPFLAGS",
    "CROSS_CURSES_INC",
    "CROSS_CURSES_LIB",
    "CVE_PRODUCT",
    "CXXFLAGS",
    "D",
    "DATA_LICENSE",
    "DATE",
    "DATETIME",
    "DEBIAN_MIRROR",
    "DEBIAN_NAMES",
    "DEBIANRDEP",
    "DEBUG_FLAGS",
    "DEBUG_OPTIMIZATION",
    "DEBUG_PREFIX_MAP",
    "DEFAULT_PREFERENCE",
    "DEFAULT_TASK_PROVIDER",
    "DEFAULTTUNE",
    "DEPCHAIN_POST",
    "DEPCHAIN_PRE",
    "DEPENDS_GOLANG",
    "DEPENDS",
    "DEPLOY_DIR_DEB",
    "DEPLOY_DIR_IMAGE",
    "DEPLOY_DIR_IPK",
    "DEPLOY_DIR_RPM",
    "DEPLOY_DIR_TAR",
    "DEPLOY_DIR_TOOLS",
    "DEPLOY_DIR",
    "DEPLOYABLE_IMAGE_TYPES",
    "DEPLOYDIR",
    "DESCRIPTION",
    "DEVSHELL_STARTDIR",
    "DEVSHELL",
    "DISABLE_STATIC",
    "DISTRO_CODENAME",
    "DISTRO_EXTRA_RDEPENDS",
    "DISTRO_EXTRA_RRECOMMENDS",
    "DISTRO_FEATURES_BACKFILL_CONSIDERED",
    "DISTRO_FEATURES_BACKFILL",
    "DISTRO_FEATURES_DEFAULT",
    "DISTRO_FEATURES_FILTER_NATIVE",
    "DISTRO_FEATURES_FILTER_NATIVESDK",
    "DISTRO_FEATURES_NATIVE",
    "DISTRO_FEATURES_NATIVESDK",
    "DISTRO_FEATURES",
    "DISTRO_NAME",
    "DISTRO_VERSION",
    "DISTROOVERRIDES",
    "DL_DIR",
    "ENABLE_BINARY_LOCALE_GENERATION",
    "ERROR_QA",
    "EXCLUDE_FROM_WORLD",
    "EXCONFIG_ARGS",
    "EXTENDPE",
    "EXTENDPKGEVER",
    "EXTENDPKGV",
    "EXTENDPRAUTO",
    "EXTRA_AUTORECONF",
    "EXTRA_IMAGE_FEATURES",
    "EXTRA_IMAGECMD",
    "EXTRA_IMAGEDEPENDS",
    "EXTRA_NATIVE_PKGCONFIG_PATH",
    "EXTRA_OECMAKE",
    "EXTRA_OECONF_PATHS",
    "EXTRA_OECONF",
    "EXTRA_OEMAKE",
    "EXTRA_OEMAKEINST",
    "EXTRA_OEMESON",
    "EXTRA_OESCONS",
    "EXTRA_STAGING_FIXMES",
    "EXTRANATIVEPATH",
    "EXTRAOPKGCONFIG",
    "FAKEROOT_QA",
    "FAKEROOTBASEENV",
    "FAKEROOTCMD",
    "FAKEROOTDIRS",
    "FAKEROOTENV",
    "FAKEROOTNOENV",
    "FEATURE_INSTALL_OPTIONAL",
    "FEATURE_INSTALL",
    "FILE_DIRNAME",
    "FILE",
    "FILES_SOLIBSDEV",
    "FILES",
    "FILESDIR",
    "FILESEXTRAPATHS",
    "FILESOVERRIDES",
    "FILESPATH",
    "FILESYSTEM_PERMS_TABLES",
    "FOSS_BASE_URL",
    "FOSS_FULL_SPDX",
    "FOSS_NO_COPYRIGHT",
    "FOSS_RECURSIVE_UNPACK",
    "FOSS_SERVER",
    "FOSS_WGET_FLAGS",
    "FULL_OPTIMIZATION",
    "GCCPIE",
    "GCCVERSION",
    "GDBVERSION",
    "GENTOO_MIRROR",
    "GLIBCVERSION",
    "GNOME_GIT",
    "GNOME_MIRROR",
    "GNU_MIRROR",
    "GNUPG_MIRROR",
    "GO_BUILD_BINDIR",
    "GO_DYNLINK",
    "GO_EXTLDFLAGS",
    "GO_IMPORT",
    "GO_INSTALL_FILTEROUT",
    "GO_INSTALL",
    "GO_LDFLAGS",
    "GO_LINKMODE",
    "GO_LINKSHARED",
    "GO_PARALLEL_BUILD",
    "GO_RPATH_LINK",
    "GO_RPATH",
    "GOCACHE",
    "GOHOSTARCH",
    "GOHOSTOS",
    "GOVERSION",
    "GPE_MIRROR",
    "GTK2DISTROFEATURES",
    "GTK3DISTROFEATURES",
    "HOMEPAGE",
    "HOST_AR_KERNEL_ARCH",
    "HOST_ARCH",
    "HOST_AS_ARCH",
    "HOST_CC_ARCH",
    "HOST_CC_KERNEL_ARCH",
    "HOST_CFLAGS",
    "HOST_CPPFLAGS",
    "HOST_CXXFLAGS",
    "HOST_EXEEXT",
    "HOST_EXTRACFLAGS",
    "HOST_GO386",
    "HOST_GOARCH",
    "HOST_GOARM",
    "HOST_GOMIPS",
    "HOST_GOOS",
    "HOST_GOTUPLE",
    "HOST_LD_ARCH",
    "HOST_LD_KERNEL_ARCH",
    "HOST_OS",
    "HOST_PREFIX",
    "HOST_SYS",
    "HOST_USER_GID",
    "HOST_USER_UID",
    "HOST_VENDOR",
    "HOSTLDFLAGS",
    "HOSTTOOLS_DIR",
    "HOSTTOOLS_NONFATAL",
    "HOSTTOOLS",
    "IMAGE_BASENAME",
    "IMAGE_CLASSES",
    "IMAGE_CMD_TAR",
    "IMAGE_FEATURES",
    "IMAGE_FSTYPES_DEBUGFS",
    "IMAGE_FSTYPES",
    "IMAGE_GEN_DEBUGFS",
    "IMAGE_INSTALL_COMPLEMENTARY",
    "IMAGE_INSTALL_DEBUGFS",
    "IMAGE_INSTALL",
    "IMAGE_LINGUAS",
    "IMAGE_LINK_NAME",
    "IMAGE_MANIFEST",
    "IMAGE_NAME_SUFFIX",
    "IMAGE_NAME",
    "IMAGE_OVERHEAD_FACTOR",
    "IMAGE_PKGTYPE",
    "IMAGE_POSTPROCESS_COMMAND",
    "IMAGE_PREPROCESS_COMMAND",
    "IMAGE_ROOTFS_ALIGNMENT",
    "IMAGE_ROOTFS_EXTRA_SPACE",
    "IMAGE_ROOTFS_SIZE",
    "IMAGE_ROOTFS",
    "IMAGE_TYPES_MASKED",
    "IMAGE_TYPES",
    "IMAGE_VERSION_SUFFIX",
    "IMGCLASSES",
    "IMGDEPLOYDIR",
    "INC_PR",
    "INHERIT_BLACKLIST",
    "INHERIT_DISTRO",
    "INHERIT",
    "INHIBIT_DEFAULT_DEPS",
    "INHIBIT_SYSROOT_STRIP",
    "INHIBIT_UPDATERCD_BBCLASS",
    "INIT_D_DIR",
    "INIT_MANAGER",
    "INITRAMFS_FSTYPES",
    "INITRAMFS_IMAGE_BUNDLE",
    "INITRAMFS_IMAGE_NAME",
    "INITRAMFS_IMAGE",
    "INITRAMFS_LINK_NAME",
    "INITRAMFS_MAXSIZE",
    "INITRAMFS_NAME",
    "INITRAMFS_TASK",
    "INITRD_IMAGE",
    "INITSCRIPT_NAME_${PN}",
    "INITSCRIPT_NAME",
    "INITSCRIPT_PACKAGES",
    "INITSCRIPT_PARAM",
    "INITSCRIPT_PARAMS_${PN}",
    "INITSCRIPT_PARAMS",
    "IPKGCONF_SDK",
    "IPKGCONF_TARGET",
    "JFFS2_ENDIANNESS",
    "JFFS2_ERASEBLOCK",
    "JFFS2_SUM_EXTRA_ARGS",
    "KBRANCH",
    "KBUILD_OUTPUT",
    "KCONF_AUDIT_LEVEL",
    "KCONF_BSP_AUDIT_LEVEL",
    "KCONFIG_CONFIG_COMMAND",
    "KEEPUIMAGE",
    "KERNEL_ALT_IMAGETYPE",
    "KERNEL_AR",
    "KERNEL_ARTIFACT_LINK_NAME",
    "KERNEL_ARTIFACT_NAME",
    "KERNEL_CC",
    "KERNEL_CLASSES",
    "KERNEL_CONFIG_COMMAND",
    "KERNEL_CONSOLE",
    "KERNEL_DEPLOYSUBDIR",
    "KERNEL_DEVICETREE_BUNDLE",
    "KERNEL_DEVICETREE",
    "KERNEL_DTB_LINK_NAME",
    "KERNEL_DTB_NAME",
    "KERNEL_EXTRA_ARGS",
    "KERNEL_EXTRA_FEATURES",
    "KERNEL_FEATURES",
    "KERNEL_FIT_LINK_NAME",
    "KERNEL_FIT_NAME",
    "KERNEL_IMAGE_LINK_NAME",
    "KERNEL_IMAGE_NAME",
    "KERNEL_IMAGEDEST",
    "KERNEL_IMAGETYPE_FOR_MAKE",
    "KERNEL_IMAGETYPE",
    "KERNEL_IMAGETYPES",
    "KERNEL_LD",
    "KERNEL_LOCALVERSION",
    "KERNEL_MODULE_AUTOLOAD",
    "KERNEL_MODULE_PACKAGE_PREFIX",
    "KERNEL_MODULE_PACKAGE_SUFFIX",
    "KERNEL_MODULE_PROBECONF",
    "KERNEL_MODULE_PROVIDE_VIRTUAL",
    "KERNEL_MODULES_META_PACKAGE",
    "KERNEL_OUTPUT_DIR",
    "KERNEL_PACKAGE_NAME",
    "KERNEL_RELEASE",
    "KERNEL_SRC_PATH",
    "KERNEL_VERSION_NAME",
    "KERNEL_VERSION_PKG_NAME",
    "KERNEL_VERSION",
    "KERNELDEPMODDEPEND",
    "KERNELORG_MIRROR",
    "KMACHINE",
    "KMETA_AUDIT",
    "KMETA",
    "LAYER_CONF_VERSION",
    "LAYERDEPENDS",
    "LAYERDIR",
    "LAYERSERIES_CORENAMES",
    "LAYERVERSION",
    "LDCONFIGDEPEND",
    "LDFLAGS",
    "LDLIBS",
    "LIBC_DEPENDENCIES",
    "LIBCEXTENSION",
    "LIBCOVERRIDE",
    "LIBCPLUSPLUS",
    "LIBS",
    "LIC_FILES_CHKSUM",
    "LICENSE_CREATE_PACKAGE",
    "LICENSE_DIRECTORY",
    "LICENSE_FILES_DIRECTORY",
    "LICENSE_PACKAGE_SUFFIX",
    "LICENSE_PATH",
    "LICENSE",
    "LICSSTATEDIR",
    "LINGUAS_INSTALL",
    "LINKER_HASH_STYLE",
    "LINUX_KERNEL_TYPE",
    "LINUX_VERSION_EXTENSION",
    "LINUX_VERSION",
    "LINUXLIBCVERSION",
    "LLVMVERSION",
    "LOCALCONF_VERSION",
    "LOCALE_SECTION",
    "LOCALE_UTF8_IS_DEFAULT",
    "LOCALE_UTF8_ONLY",
    "LOCALEBASEPN",
    "LOG_DIR",
    "LOGFIFO",
    "LUA_TARGET_OS",
    "MACHINE_ARCH",
    "MACHINE_ESSENTIAL_EXTRA_RDEPENDS",
    "MACHINE_ESSENTIAL_EXTRA_RRECOMMENDS",
    "MACHINE_EXTRA_RDEPENDS",
    "MACHINE_EXTRA_RRECOMMENDS",
    "MACHINE_FEATURES_BACKFILL_CONSIDERED",
    "MACHINE_FEATURES_BACKFILL",
    "MACHINE_FEATURES",
    "MACHINE_HWCODECS",
    "MACHINE_TASK_PROVIDER",
    "MACHINEOVERRIDES",
    "MAINTAINER",
    "METADATA_BRANCH",
    "METADATA_REVISION",
    "MIN_BTRFS_SIZE",
    "MIN_F2FS_SIZE",
    "MINOR_VERSION",
    "MIRRORS",
    "MLPREFIX",
    "MODULE_TARBALL_DEPLOY",
    "MODULE_TARBALL_LINK_NAME",
    "MODULE_TARBALL_NAME",
    "MULTI_PROVIDER_WHITELIST",
    "MULTILIB_CHECK_FILE",
    "MULTILIB_TEMP_ROOTFS",
    "MULTILIB_VARIANTS",
    "MULTILIBRE_ALLOW_REP",
    "MULTIMACH_TARGET_SYS",
    "NATIVE_PACKAGE_PATH_SUFFIX",
    "NATIVELSBSTRING",
    "NO_RECOMMENDATIONS",
    "NO32LIBS",
    "NPM_ARCH",
    "NPM_EXTRA_ARGS",
    "NPM_LIBDIR",
    "NPM_TAR_FILENAME",
    "OE_DEL",
    "OE_EXTRA_IMPORTS",
    "OE_FEATURES",
    "OE_IMPORTED",
    "OE_IMPORTS",
    "OE_INIT_ENV_SCRIPT",
    "OE_TERMINAL_EXPORTS",
    "OE_TERMINAL",
    "OECMAKE_RPATH",
    "OECMAKE_SOURCEPATH",
    "OEINCLUDELOGS",
    "OELAYOUT_ABI",
    "OES_BITBAKE_CONF",
    "OLDEST_KERNEL",
    "OPKG_ARGS",
    "OPKG_POSTPROCESS_COMMANDS",
    "OPKG_PREPROCESS_COMMANDS",
    "OPKGBUILDCMD",
    "OPKGLIBDIR",
    "OVERRIDES",
    "P",
    "PACKAGE_ARCH",
    "PACKAGE_ARCHS",
    "PACKAGE_BEFORE_PN",
    "PACKAGE_CLASSES",
    "PACKAGE_DEBUG_SPLIT_STYLE",
    "PACKAGE_DEPENDS",
    "PACKAGE_EXCLUDE",
    "PACKAGE_EXTRA_ARCHS",
    "PACKAGE_INSTALL_ATTEMPTONLY",
    "PACKAGE_PREPROCESS_FUNCS",
    "PACKAGE_WRITE_DEPS",
    "PACKAGEBUILDPKGD",
    "PACKAGECONFIG_CONFARGS",
    "PACKAGECONFIG_NUMA",
    "PACKAGECONFIG",
    "PACKAGEFUNCS",
    "PACKAGEINDEXDEPS",
    "PACKAGES_DYNAMIC",
    "PACKAGES",
    "PACKAGESPLITFUNCS",
    "PACKAGEVARS",
    "PARALLEL_MAKE",
    "PARALLEL_MAKEINST",
    "PATCH_GIT_USER_EMAIL",
    "PATCH_GIT_USER_NAME",
    "PATCHDEPENDENCY",
    "PATCHRESOLVE",
    "PATCHTOOL",
    "PE",
    "PERSISTENT_DIR",
    "PF",
    "PHP_LIBDIR",
    "PID",
    "PKGD",
    "PKGDATA_DIR",
    "PKGDATA_VARS",
    "PKGDEST",
    "PKGDESTWORK",
    "PKGE",
    "PKGMLTRIPLETS",
    "PKGR",
    "PKGTRIPLETS",
    "PKGV",
    "PKGWRITEDIRIPK",
    "PN",
    "POKY_BBLAYERS_CONF_VERSION",
    "POKY_DEFAULT_DISTRO_FEATURES",
    "POKY_DEFAULT_EXTRA_RDEPENDS",
    "POKY_DEFAULT_EXTRA_RRECOMMENDS",
    "POKYQEMUDEPS",
    "POPULATE_SDK_POST_HOST_COMMAND",
    "POPULATESYSROOTDEPS",
    "POSTINST_INTERCEPT_CHECKSUMS",
    "POSTINST_INTERCEPTS_DIR",
    "POSTINST_INTERCEPTS_PATHS",
    "POSTINST_INTERCEPTS",
    "POSTINST_LOGFILE",
    "PPPD_VERSION",
    "PR",
    "PRAUTO",
    "PRAUTOINX",
    "PREFERRED_PROVIDER",
    "PREFERRED_PROVIDERS",
    "PREFERRED_VERSION",
    "PREFIX",
    "PREMIRRORS",
    "PREPROCESS_RELOCATE_DIRS",
    "PRIORITY",
    "PROVIDES",
    "PRSERV_HOST",
    "PSEUDO_LOCALSTATEDIR",
    "PSEUDO_PASSWD",
    "PSEUDO_SYSROOT",
    "PTEST_BINDIR_PKGD_PATH",
    "PTEST_BINDIR",
    "PTEST_BUILD_HOST_FILES",
    "PTEST_BUILD_HOST_PATTERN",
    "PTEST_ENABLED",
    "PTEST_PATH",
    "PV",
    "PYTHON_ABI",
    "PYTHON_BASEVERSION",
    "PYTHON_BINABI",
    "PYTHON_DIR",
    "PYTHON_PN",
    "PYTHON_SITE_PACKAGES",
    "PYTHON_SITEPACKAGES_DIR",
    "PYTHON",
    "QA_LOGFILE",
    "QA_SANE",
    "QB_AUDIO_DRV",
    "QB_AUDIO_OPT",
    "QB_CPU_KVM",
    "QB_CPU",
    "QB_DEFAULT_FSTYPE",
    "QB_DEFAULT_KERNEL",
    "QB_DRIVE_TYPE",
    "QB_KERNEL_CMDLINE_APPEND",
    "QB_MEM",
    "QB_NETWORK_DEVICE",
    "QB_OPT_APPEND",
    "QB_SERIAL_OPT",
    "QB_SYSTEM_NAME",
    "QEMU_OPTIONS",
    "QEMUVERSION",
    "QUILTRCFILE",
    "RDEPENDS",
    "REAL_MULTIMACH_TARGET_SYS",
    "RECIPE_MAINTAINER",
    "RECIPE_SYSROOT_NATIVE",
    "RECIPE_SYSROOT",
    "RECIPERDEPTASK",
    "REMOVE_LIBTOOL_LA",
    "REQUIRED_POKY_BBLAYERS_CONF_VERSION",
    "RM_WORK_BUILD_WITHOUT",
    "RM_WORK_EXCLUDE_ITEMS",
    "RM_WORK_EXCLUDE",
    "RMINITDIR",
    "ROOT_HOME",
    "ROOTFS_BOOTSTRAP_INSTALL",
    "ROOTFS_PKGMANAGE",
    "ROOTFS_RO_UNNEEDED",
    "RPMDEPS",
    "RPROVIDES",
    "RRECOMMENDS",
    "RUNNABLE_IMAGE_TYPES",
    "RUNNABLE_MACHINE_PATTERNS",
    "RUNTIME",
    "S",
    "SANITY_ABIFILE",
    "SANITY_BBLAYERCONF_SAMPLE",
    "SANITY_DIFF_TOOL",
    "SANITY_LOCALCONF_SAMPLE",
    "SANITY_REQUIRED_UTILITIES",
    "SANITY_SITECONF_SAMPLE",
    "SANITY_TESTED_DISTROS",
    "SANITY_VERSION",
    "SAVANNAH_GNU_MIRROR",
    "SAVANNAH_NONGNU_MIRROR",
    "SDK_ARCH",
    "SDK_ARCHIVE_CMD",
    "SDK_ARCHIVE_DEPENDS",
    "SDK_ARCHIVE_TYPE",
    "SDK_AS_ARCH",
    "SDK_CC_ARCH",
    "SDK_DEPENDS",
    "SDK_DEPLOY",
    "SDK_DIR",
    "SDK_EXT_HOST_MANIFEST",
    "SDK_EXT_TARGET_MANIFEST",
    "SDK_EXT_TYPE",
    "SDK_EXT",
    "SDK_HOST_MANIFEST",
    "SDK_INCLUDE_NATIVESDK",
    "SDK_INCLUDE_PKGDATA",
    "SDK_INCLUDE_TOOLCHAIN",
    "SDK_INHERIT_BLACKLIST",
    "SDK_INSTALL_TARGETS",
    "SDK_LD_ARCH",
    "SDK_LOCAL_CONF_BLACKLIST",
    "SDK_LOCAL_CONF_WHITELIST",
    "SDK_NAME_PREFIX",
    "SDK_NAME",
    "SDK_OLDEST_KERNEL",
    "SDK_OS",
    "SDK_OUTPUT",
    "SDK_PACKAGE_ARCHS",
    "SDK_PACKAGING_COMMAND",
    "SDK_PACKAGING_FUNC",
    "SDK_POST_INSTALL_COMMAND",
    "SDK_PRE_INSTALL_COMMAND",
    "SDK_PREFIX",
    "SDK_RDEPENDS",
    "SDK_RECRDEP_TASKS",
    "SDK_RELOCATE_AFTER_INSTALL",
    "SDK_SYS",
    "SDK_TARGET_MANIFEST",
    "SDK_TARGETS",
    "SDK_TITLE",
    "SDK_UPDATE_URL",
    "SDK_VENDOR",
    "SDK_VERSION",
    "SDKDEPLOYDIR",
    "SDKEXTDEPLOYDIR",
    "SDKEXTPATH",
    "SDKGCCVERSION",
    "SDKIMAGE_FEATURES",
    "SDKIMAGE_INSTALL_COMPLEMENTARY",
    "SDKIMAGE_LINGUAS",
    "SDKMACHINE",
    "SDKPATH",
    "SDKPATHNATIVE",
    "SDKPKGSUFFIX",
    "SDKTARGETSYSROOT",
    "SDKTAROPTS",
    "SDKUSE_NLS",
    "SECTION",
    "SECURITY_CFLAGS",
    "SECURITY_LDFLAGS",
    "SECURITY_NO_PIE_CFLAGS",
    "SECURITY_NOPIE_CFLAGS",
    "SECURITY_PIE_CFLAGS",
    "SECURITY_STACK_PROTECTOR",
    "SECURITY_STRINGFORMAT",
    "SECURITY_X_LDFLAGS",
    "SELECTED_OPTIMIZATION",
    "SERIAL_CONSOLE",
    "SERIAL_CONSOLES",
    "SHLIBSDIRS",
    "SHLIBSWORKDIR",
    "SIGGEN_EXCLUDE_SAFE_RECIPE_DEPS",
    "SIGGEN_EXCLUDERECIPES_ABISAFE",
    "SIGGEN_LOCKEDSIGS_SSTATE_EXISTS_CHECK",
    "SIGGEN_LOCKEDSIGS_TASKSIG_CHECK",
    "SITE_CONF_VERSION",
    "SITECONFIG_SYSROOTCACHE",
    "SITEINFO_BITS",
    "SITEINFO_ENDIANNESS",
    "SOLIBS",
    "SOLIBSDEV",
    "SORT_PASSWD_POSTPROCESS_COMMAND",
    "SOURCEFORGE_MIRROR",
    "SPDX_MANIFEST_DIR",
    "SPDX_TEMP_DIR",
    "SPDX_VERSION",
    "SPECIAL_PKGSUFFIX",
    "SPLASH",
    "SRC_DISTRIBUTE_LICENSES",
    "SRC_URI",
    "SRCDATE",
    "SRCPV",
    "SRCREV_FORMAT",
    "SRCREV",
    "SRCTREECOVEREDTASKS",
    "SSH_AGENT_PID",
    "SSH_AUTH_SOCK",
    "SSTATE_ARCHS",
    "SSTATE_DIR",
    "SSTATE_DUPWHITELIST",
    "SSTATE_EXCLUDEDEPS_SYSROOT",
    "SSTATE_EXTRAPATH",
    "SSTATE_EXTRAPATHWILDCARD",
    "SSTATE_HASHEQUIV_METHOD",
    "SSTATE_HASHEQUIV_REPORT_TASKDATA",
    "SSTATE_MANFILEPREFIX",
    "SSTATE_MANIFESTS",
    "SSTATE_MANMACH",
    "SSTATE_PATHSPEC",
    "SSTATE_PKG",
    "SSTATE_PKGARCH",
    "SSTATE_PKGNAME",
    "SSTATE_PKGSPEC",
    "SSTATE_PRUNE_OBSOLETEWORKDIR",
    "SSTATE_SCAN_CMD_NATIVE",
    "SSTATE_SCAN_CMD",
    "SSTATE_SCAN_FILES",
    "SSTATE_SIG_KEY",
    "SSTATE_SIG_PASSPHRASE",
    "SSTATE_SWSPEC",
    "SSTATE_VERIFY_SIG",
    "SSTATE_VERSION",
    "SSTATECREATEFUNCS",
    "SSTATEPOSTCREATEFUNCS",
    "SSTATEPOSTINSTFUNCS",
    "SSTATEPOSTUNPACKFUNCS",
    "SSTATEPREINSTFUNCS",
    "SSTATETASKS",
    "STAGING_BASE_LIBDIR_NATIVE",
    "STAGING_BASELIBDIR",
    "STAGING_BINDIR_CROSS",
    "STAGING_BINDIR_NATIVE",
    "STAGING_BINDIR_TOOLCHAIN",
    "STAGING_BINDIR",
    "STAGING_DATADIR_NATIVE",
    "STAGING_DATADIR",
    "STAGING_DIR_HOST",
    "STAGING_DIR_NATIVE",
    "STAGING_DIR_TARGET",
    "STAGING_DIR",
    "STAGING_ETCDIR_NATIVE",
    "STAGING_EXECPREFIXDIR",
    "STAGING_FIRMWARE_DIR",
    "STAGING_INCDIR_NATIVE",
    "STAGING_INCDIR",
    "STAGING_KERNEL_BUILDDIR",
    "STAGING_KERNEL_DIR",
    "STAGING_LIBDIR_NATIVE",
    "STAGING_LIBDIR",
    "STAGING_LIBEXECDIR_NATIVE",
    "STAGING_LIBEXECDIR",
    "STAGING_LOADER_DIR",
    "STAGING_SBINDIR_NATIVE",
    "STAMP",
    "STAMPCLEAN",
    "STAMPS_DIR",
    "SUMMARY",
    "SWIG_FEATURES",
    "SYSROOT_DESTDIR",
    "SYSROOT_DIR",
    "SYSROOT_DIRS_BLACKLIST",
    "SYSROOT_DIRS_NATIVE",
    "SYSROOT_DIRS",
    "SYSROOT_PREPROCESS_FUNCS",
    "SYSTEMD_AUTO_ENABLE_${PN}",
    "SYSTEMD_AUTO_ENABLE",
    "SYSTEMD_DEFAULT_TARGET",
    "SYSTEMD_PACKAGES",
    "T",
    "TARGET_AR_KERNEL_ARCH",
    "TARGET_AS_ARCH",
    "TARGET_CC_ARCH",
    "TARGET_CC_KERNEL_ARCH",
    "TARGET_CFLAGS",
    "TARGET_CPPFLAGS",
    "TARGET_CXXFLAGS",
    "TARGET_FPU",
    "TARGET_GO386",
    "TARGET_GOARCH",
    "TARGET_GOARM",
    "TARGET_GOMIPS",
    "TARGET_GOOS",
    "TARGET_GOTUPLE",
    "TARGET_LD_ARCH",
    "TARGET_LD_KERNEL_ARCH",
    "TARGET_LDFLAGS",
    "TARGET_LINK_HASH_STYLE",
    "TARGET_OS",
    "TARGET_PREFIX",
    "TARGET_SYS",
    "TARGET_VENDOR",
    "TCLIBC",
    "TCLIBCAPPEND",
    "TCMODE",
    "TERMINFO",
    "THISDIR",
    "TIME",
    "TMPDIR",
    "TOOLCHAIN_HOST_TASK_ATTEMPTONLY",
    "TOOLCHAIN_HOST_TASK",
    "TOOLCHAIN_OPTIONS",
    "TOOLCHAIN_OUTPUTNAME",
    "TOOLCHAIN_SHAR_EXT_TMPL",
    "TOOLCHAIN_SHAR_REL_TMPL",
    "TOOLCHAIN_TARGET_TASK_ATTEMPTONLY",
    "TOOLCHAIN_TARGET_TASK",
    "TOOLCHAIN",
    "TOOLCHAINEXT_OUTPUTNAME",
    "TOPDIR",
    "TOPDIR",
    "TRANSLATED_TARGET_ARCH",
    "TUNE_ARCH",
    "TUNE_ASARGS",
    "TUNE_CCARGS",
    "TUNE_FEATURES",
    "TUNE_LDARGS",
    "TUNE_PKGARCH",
    "TZ_PACKAGES",
    "UBI_VOLNAME",
    "UBOOT_ENTRYPOINT",
    "UBOOT_LOADADDRESS",
    "UBOOT_MACHINE",
    "UNINATIVE_DLDIR",
    "UNINATIVE_LOADER",
    "UNINATIVE_MAXGLIBCVERSION",
    "UNINATIVE_STAGING_DIR",
    "UNINATIVE_TARBALL",
    "UNINATIVE_URL",
    "UNKNOWN_CONFIGURE_WHITELIST",
    "UNWINDLIB",
    "UPDALTVARS",
    "UPDATERCD",
    "UPDATERCPN",
    "UPSTREAM_CHECK_COMMITS",
    "UPSTREAM_CHECK_GITTAGREGEX",
    "UPSTREAM_CHECK_REGEX",
    "UPSTREAM_CHECK_URI",
    "USE_DEPMOD",
    "USE_DEVFS",
    "USE_NLS",
    "USER_CLASSES",
    "USING_WIC",
    "USRBINPATH",
    "UVESA_MODE",
    "VOLATILE_LOG_DIR",
    "WARN_QA",
    "WARN_TO_ERROR_QA",
    "WIC_CREATE_EXTRA_ARGS",
    "WICVARS",
    "WKS_FILE_CHECKSUM",
    "WKS_FILE_DEPENDS_BOOTLOADERS",
    "WKS_FILE_DEPENDS_DEFAULT",
    "WKS_FILE_DEPENDS",
    "WKS_FILE",
    "WKS_FILES",
    "WKS_FULL_PATH",
    "WKS_SEARCH_PATH",
    "WORKDIR_PKGDATA",
    "WORKDIR",
    "X86ARCH32",
    "X86ARCH64",
    "XAUTHORITY",
    "XLIBS_MIRROR",
    "XORG_MIRROR",
    "XSERVER",
    "XZ_COMPRESSION_LEVEL",
    "XZ_DEFAULTS",
    "XZ_INTEGRITY_CHECK",
    "ZIP_COMPRESSION_LEVEL"
]


def get_known_vars():
    return __get_const_and_rulefile(KNOWN_VARS, "known_vars")



import argparse
import os
import sys
import json

from oelint_adv.cls_rule import load_rules
from oelint_adv.cls_stash import Stash
from oelint_adv.color import set_color
from oelint_adv.rule_file import set_rulefile, set_constantfile

sys.path.append(os.path.abspath(os.path.join(__file__, "..")))


def create_argparser():
    parser = argparse.ArgumentParser(
        description='Advanced OELint - Check bitbake recipes against OECore styleguide')
    parser.add_argument("--suppress", default=[],
                        action="append", help="Rules to suppress")
    parser.add_argument("--output", default=sys.stderr,
                        help="Where to flush the findings (default: stderr)")
    parser.add_argument("--fix", action="store_true", default=False,
                        help="Automatically try to fix the issues")
    parser.add_argument("--nobackup", action="store_true", default=False,
                        help="Don't create backup file when auto fixing")
    parser.add_argument("--addrules", nargs="+", default=[],
                        help="Additional non-default rulessets to add")
    parser.add_argument("--rulefile", default=None,
                        help="Rulefile")
    parser.add_argument("--constantfile", default=None, help="Constantfile")
    parser.add_argument("--color", action="store_true", default=False,
                        help="Add color to the output based on the severity")
    parser.add_argument("files", nargs='+', help="File to parse")

    args = parser.parse_args()

    if args.rulefile:
        try:
            with open(args.rulefile) as i:
                set_rulefile(json.load(i))
        except (FileNotFoundError, json.JSONDecodeError):
            raise argparse.ArgumentTypeError("'rulefile' is not a valid file")
    
    if args.constantfile:
        try:
            with open(args.constantfile) as i:
                set_constantfile(json.load(i))
        except (FileNotFoundError, json.JSONDecodeError):
            raise argparse.ArgumentTypeError("'constantfile' is not a valid file")

    if args.color:
        set_color(True)
    return args


if __name__ == '__main__':
    args = create_argparser()
    rules = [x for x in load_rules(
        add_rules=args.addrules) if str(x) not in args.suppress]
    print("Loaded rules: {}".format(",".join(sorted([str(x) for x in rules]))))
    stash = Stash()
    issues = []
    fixedfiles = []
    for f in args.files:
        try:
            stash.AddFile(f)
        except FileNotFoundError as e:
            print("Can't open/read: {}".format(e))

    for f in list(set(stash.GetRecipes() + stash.GetLoneAppends())):
        for r in rules:
            if not r.OnAppend and f.endswith(".bbappend"):
                continue
            if r.OnlyAppend and not f.endswith(".bbappend"):
                continue
            if args.fix:
                fixedfiles += r.fix(f, stash)
            issues += r.check(f, stash)
    fixedfiles = list(set(fixedfiles))
    for f in fixedfiles:
        _items = [f] + stash.GetLinksForFile(f)
        for i in _items:
            items = sorted(stash.GetItemsFor(filename=i, nolink=True), key=lambda x: x.Line)
            if not args.nobackup:
                os.rename(i, i + ".bak")
            with open(i, "w") as o:
                o.write("".join([x.Raw for x in items]))
                print("{}:{}:{}".format(os.path.abspath(i),
                                        "debug", "Applied automatic fixes"))

    issues = sorted(list(set(issues)))

    if args.output != sys.stderr:
        args.output = open(args.output, "w")
    args.output.write("\n".join(issues) + "\n")
    if args.output != sys.stderr:
        args.output.close()
    sys.exit(len(issues))



from oelint_adv.cls_item import Include
from oelint_adv.cls_rule import Rule


class FileIncludeVsRequire(Rule):
    def __init__(self):
        super().__init__(id="oelint.file.requireinclude",
                         severity="warning",
                         message="Use 'require {FILE}' instead of 'include {FILE}'")

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file,
                                      classifier=Include.CLASSIFIER):
            if item.Statement == "include":
                res += self.finding(item.Origin, item.Line,
                                    self.Msg.replace("{FILE}", item.IncName))
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarNativeFilename(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.nativefilename",
                         severity="warning",
                         message="native-recipe-files should include '-native' in file name")

    def check(self, _file, stash):
        res = []
        items = [x for x in
                 stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue="inherit")
                 if x.VarValueStripped == "native"]
        if any(items):
            if _file.find("-native") == -1:
                res += self.finding(_file, 0)
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarQuoted(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.pkgspecific",
                         severity="error",
                         message="Variable {VAR} is package-specific and therefore it should be {VAR}_${PN}")

    def check(self, _file, stash):
        res = []
        needles = ['RDEPENDS', 'RRECOMMENDS', 'RSUGGESTS', 'RCONFLICTS', 'RPROVIDES', 'RREPLACES',
                   'FILES', 'pkg_preinst', 'pkg_postinst', 'pkg_prerm', 'pkg_postrm', 'ALLOW_EMPTY']
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if i.VarName == "inherit":
                continue
            if i.VarName in needles:
                res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.replace("{VAR}", i.VarName))
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarSectionLowercase(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.doublemodify",
                         severity="error",
                         message="Multiple modifiers of append/prepend/remove/+= found in one operation")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in [x for x in items]:
            if len(i.AppendOperation()) > 1:
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarPnUsageDiscouraged(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.pnusagediscouraged",
                         severity="warning",
                         message="Variable shouldn't contain ${PN} or ${BPN}")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        needles = ["SUMMARY", "HOMEPAGE", "BUGTRACKER", "DESCRIPTION"]
        for i in [x for x in items if x.VarName in needles]:
            if "${PN}" in i.VarValue or "${BPN}" in i.VarValue:
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_item import Function
from oelint_adv.cls_item import TaskAdd
from oelint_adv.cls_rule import Rule
from oelint_adv.const_func import FUNC_ORDER


class TaskAddNoTaskBody(Rule):
    def __init__(self):
        super().__init__(id="oelint.task.addnotaskbody",
                         severity="warning",
                         message="The added task '{FUNC}' is not existing or has no body")

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=TaskAdd.CLASSIFIER):
            if item.FuncName in FUNC_ORDER:
                # not for builtin types
                continue
            if item.FuncName in [x.replace("do_", "", 1) for x in FUNC_ORDER]:
                # not for builtin types - probed for missing prefix
                continue
            _ta = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER,
                                    attribute="FuncName")
            _filt = [x for x in _ta or [] if not isinstance(x, str) and x.FuncName == item.FuncName]
            _filt += [x for x in _ta or [] if not isinstance(x, str) and x.FuncName == "do_" + item.FuncName]
            if not any(_filt):
                res += self.finding(item.Origin, item.InFileLine, self.Msg.replace("{FUNC}", item.FuncName))
            elif not any([x for x in _filt if x.FuncBodyStripped]):
                res += self.finding(item.Origin, item.InFileLine, self.Msg.replace("{FUNC}", item.FuncName))
        return res



from oelint_adv.cls_rule import Rule


class NoCommentsTrailing(Rule):
    def __init__(self):
        super().__init__(id="oelint.comments.notrailing",
                         severity="error",
                         message="Comments shall be put on seperate lines")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        for i in items:
            if i.Raw:
                for line in i.Raw.split("\n"):
                    line = line.strip()
                    if "#" in line and line.find("#") > 0:
                        res += self.finding(i.Origin, i.InFileLine)
        return res



import os

from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.helper_files import get_files


class FilePatchIsSignedOff(Rule):
    def __init__(self):
        super().__init__(id="oelint.file.patchsignedoff",
                         severity="warning",
                         message="Patch '{FILE}' should should contain a Signed-Off entry")

    def check(self, _file, stash):
        res = []
        items = get_files(stash, _file, "*.patch")
        _items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")
        for i in items:
            with open(i) as _input:
                try:
                    content = _input.readlines()
                    if not any([x for x in content if x.startswith("Signed-off-by: ")]):
                        # Find matching SRC_URI assignment
                        _assign = [x for x in _items if x.VarValue.find(
                            os.path.basename(i)) != -1]
                        if any(_assign):
                            res += self.finding(_assign[0].Origin,
                                                _assign[0].InFileLine,
                                                self.Msg.replace("{FILE}", os.path.basename(i)))
                except UnicodeDecodeError:
                    pass
        return res



import setuptools

import subprocess
_long_description = "See https://github.com/priv-kweihmann/oelint-adv for documentation"
_long_description_content_type = "text/plain"
try:
    _long_description = subprocess.check_output(
        ["pandoc", "--from", "markdown", "--to", "rst", "README.md"]).decode("utf-8")
    _long_description_content_type = "text/x-rst"
except (subprocess.CalledProcessError, FileNotFoundError):
    pass

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="oelint_adv",
    version="1.15.0",
    author="Konrad Weihmann",
    author_email="kweihmann@outlook.com",
    description="Advanced bitbake-recipe linter",
    long_description=_long_description,
    long_description_content_type=_long_description_content_type,
    url="https://github.com/priv-kweihmann/oelint-adv",
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "oelint-adv = oelint_adv.__main__:main",
        ]
    },
    install_requires=requirements,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Quality Assurance",
    ],
)



from oelint_adv.cls_item import Function
from oelint_adv.cls_rule import Rule


class TaskInstallNoMkdir(Rule):
    def __init__(self):
        super().__init__(id="oelint.task.nomkdir",
                         severity="error",
                         message="'mkdir' shall not be used in do_install. Use 'install'")

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER):
            if item.FuncName.startswith("do_install"):
                if item.FuncBody.find(" mkdir ") != -1:
                    res += self.finding(item.Origin, item.InFileLine)
        return res



import os
import re

from oelint_adv.parser import get_items


class Stash():

    def __init__(self):
        self.__list = []

    def AddFile(self, _file, lineOffset=0, forcedLink=None):
        print("Parsing {}".format(_file))
        res = get_items(self, _file, lineOffset=lineOffset)
        if forcedLink:
            for item in [x for x in res if x.Origin == _file]:
                item.AddLink(forcedLink)
        # Match bbappends to bbs
        if _file.endswith(".bbappend"):
            bn_this = os.path.basename(_file).replace(
                ".bbappend", "").replace("%", ".*")
            for item in self.__list:
                if re.match(bn_this, os.path.basename(item.Origin).replace(".bb", "")):
                    item.AddLink(_file)
                    for r in res:
                        r.AddLink(item.Origin)
        self.__list += res
        return res

    def GetRecipes(self):
        return list(set([x.Origin for x in self.__list if x.Origin.endswith(".bb")]))

    def GetLoneAppends(self):
        __linked_appends = []
        __appends = []
        for x in self.__list:
            if x.Origin.endswith(".bbappend"):
                __appends.append(x)
            else:
                __linked_appends += x.Links
        return list(set([x.Origin for x in __appends if x not in __linked_appends]))

    def __is_linked_to(self, item, filename, nolink=False):
        return (filename in item.Links and not nolink) or filename == item.Origin

    def __get_items_by_file(self, items, filename, nolink=False):
        if not filename:
            return items
        return [x for x in items if self.__is_linked_to(x, filename, nolink=nolink)]

    def __get_items_by_classifier(self, items, classifier):
        if not classifier:
            return items
        return [x for x in items if x.CLASSIFIER == classifier]

    def __get_items_by_attribute(self, items, attname, attvalue):
        if not attname:
            return items
        # v is a list
        res = [x for x in items if attname in x.GetAttributes().keys()]
        if attvalue:
            res = [x for x in res if (attname in x.GetAttributes(
            ).keys() and x.GetAttributes()[attname] == attvalue)]
        return res

    def GetLinksForFile(self, filename):
        if not filename:
            return []
        return [x.Origin for x in self.__get_items_by_file(self.__list, filename) if x.Origin != filename]

    def GetItemsFor(self, filename=None, classifier=None, attribute=None, attributeValue=None, nolink=False):
        res = self.__list
        res = self.__get_items_by_file(res, filename, nolink=nolink)
        res = self.__get_items_by_classifier(res, classifier)
        res = self.__get_items_by_attribute(res, attribute, attributeValue)
        return list(set(res))



import re

from oelint_adv.cls_item import Function
from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarPnBpnUsage(Rule):
    def __init__(self):
        super().__init__(id="oelint.func.machinespecific",
                         severity="error",
                         message="'{}' is set machine specific ['{}'], but a matching COMPATIBLE_MACHINE entry is missing")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER,
                                  attribute=Function.ATTR_FUNCNAME)
        for i in items:
            _machine = i.GetMachineEntry()
            if not _machine:
                continue
            _comp = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                      attribute=Variable.ATTR_VAR, attributeValue="COMPATIBLE_MACHINE")
            if not any(_comp):
                res += self.finding(i.Origin, i.InFileLine,
                                    override_msg=self.Msg.format(i.FuncName, _machine))
                continue
            _vals = [x.VarValueStripped.lstrip(
                "|") for x in _comp if x.VarValueStripped]
            if not any(re.match(v, _machine) or (_machine == "qemuall" and "qemu" in v) for v in _vals):
                res += self.finding(i.Origin, i.InFileLine,
                                    override_msg=self.Msg.format(i.FuncName, _machine))
        return res



import re

from oelint_adv.cls_item import Include
from oelint_adv.cls_rule import Rule


class VarMultiInclude(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.multiinclude",
                         severity="warning",
                         message="'{INC}' is included multiple times")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Include.CLASSIFIER)
        keys = []
        for i in items:
            keys += [x.strip() for x in re.split(r"\s|,", i.IncName) if x]
        for key in list(set(keys)):
            _i = [x for x in items if x.IncName.find(key) != -1]
            if len(_i) > 1:
                res += self.finding(_i[-1].Origin, _i[-1].InFileLine,
                                    self.Msg.replace("{INC}", key))
        return res



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_adv.helper_files import get_scr_components


class VarPnBpnUsage(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.pnbpnusage",
                         severity="error",
                         message="${BPN} should be used instead of ${PN}")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        needles = ["SRC_URI"]
        for i in [x for x in items if x.VarName in needles]:
            _comp = get_scr_components(i.VarValue)
            if "${PN}" in _comp["src"]:
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_adv.cls_item import Variable, Comment

from collections import OrderedDict


class VarsPathHardcode(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.pathhardcode",
                         severity="warning",
                         message="<FOO>")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        _map = OrderedDict({
            "/usr/lib/systemd/user": "${systemd_user_unitdir}",
            "/lib/systemd/system": "${systemd_system_unitdir}",
            "/usr/share/doc": "${docdir}",
            "/usr/share/info": "${infodir}",
            "/usr/share/man": "${mandir}",
            "/usr/libexec": "${libexecdir}",
            "/lib/systemd": "${systemd_unitdir}",
            "/usr/lib": "${libdir}",
            "/usr/bin": "${bindir}",
            "/usr/share": "${datadir}",
            "/usr/include": "${includedir}",
            "/var": "${localstatedir}",
            "/lib": "${nonarch_base_libdir}",
            "/usr/sbin": "${sbindir}",
            "/srv": "${servicedir}",
            "/com": "${sharedstatedir}",
            "/etc": "${sysconfdir}",
        })
        for i in items:
            if isinstance(i, Variable) and \
               i.VarName in ["SUMMARY", "DESCRIPTION", "HOMEPAGE", "AUTHOR", "BUGTRACKER", "FILES", "FILES_${PN}"]:
                continue
            if isinstance(i, Comment):
                continue
            _matches = []
            for k, v in _map.items():
                for pre in ["'", "\"", " ", "${D}", "="]:
                    for ext in [" ", "/", "\"", "*"]:
                        if pre + k + ext in i.Raw and k not in _matches:
                            res += self.finding(i.Origin, i.InFileLine,
                                                "'{}' should be used instead of '{}'".format(v, k))
                            _parts = [x for x in k.split("/") if x]
                            for x in range(0, len(_parts)):
                                _matches.append("/" + "/".join(_parts[0:x]))
        return res



FUNC_ORDER = [
    "do_fetch",
    "do_unpack",
    "do_patch",
    "do_configure",
    "do_compile",
    "do_install",
    "do_populate_sysroot",
    "do_build",
    "do_package"
]

KNOWN_FUNCS = [
    "do_addto_recipe_sysroot",
    "do_allpackagedata",
    "do_ar_configured",
    "do_ar_original",
    "do_ar_patched",
    "do_ar_recipe",
    "do_assemble_fitimage",
    "do_assemble_fitimage_initramfs",
    "do_bootimg",
    "do_build",
    "do_build_native_sysroot",
    "do_build_sysroot",
    "do_build_target_sysroot",
    "do_build_without_rm_work",
    "do_bundle_files",
    "do_bundle_initramfs",
    "do_ccmake",
    "do_ccmake_diffconfig",
    "do_check",
    "do_checkuri",
    "do_clean",
    "do_cleanall",
    "do_cleanccache",
    "do_cleansstate",
    "do_collect_packagedata",
    "do_compile",
    "do_compile_kernelmodules",
    "do_compile_ptest_base",
    "do_configure",
    "do_configure_ptest_base",
    "do_convert_crlf_to_lf",
    "do_create_extlinux_config",
    "do_create_manifest",
    "do_create_rdepends_inc",
    "do_cve_check",
    "do_deploy",
    "do_deploy_archives",
    "do_deploy_archives_setscene",
    "do_deploy_setscene",
    "do_deploy_source_date_epoch",
    "do_deploy_source_date_epoch_setscene",
    "do_devpyshell",
    "do_devshell",
    "do_diffconfig",
    "do_dumpdata",
    "do_extra_symlinks",
    "do_fetch",
    "do_gcc_stash_builddir",
    "do_gcc_stash_builddir_setscene",
    "do_generate_content",
    "do_generate_toolchain_file",
    "do_get_public_keys",
    "do_image",
    "do_image_complete",
    "do_image_complete_setscene",
    "do_image_qa",
    "do_image_qa_setscene",
    "do_install",
    "do_install_ptest_base",
    "do_install_ptest_perl",
    "do_kernel_checkout",
    "do_kernel_configcheck",
    "do_kernel_configme",
    "do_kernel_link_images",
    "do_kernel_metadata",
    "do_kernel_version_sanity_check",
    "do_lint",
    "do_listtasks",
    "do_locked_sigs",
    "do_menuconfig",
    "do_mkimage",
    "do_multilib_install",
    "do_package",
    "do_package_index",
    "do_package_qa",
    "do_package_qa_setscene",
    "do_package_setscene",
    "do_package_write_deb",
    "do_package_write_deb_setscene",
    "do_package_write_ipk",
    "do_package_write_ipk_setscene",
    "do_package_write_rpm",
    "do_package_write_rpm_setscene",
    "do_package_write_tar",
    "do_packagedata",
    "do_packagedata_setscene",
    "do_patch",
    "do_populate_cve_db",
    "do_populate_ide_support",
    "do_populate_lic",
    "do_populate_lic_deploy",
    "do_populate_lic_setscene",
    "do_populate_sdk",
    "do_populate_sdk_ext",
    "do_populate_sysroot",
    "do_populate_sysroot_setscene",
    "do_preconfigure",
    "do_prepare_copyleft_sources",
    "do_prepare_recipe_sysroot",
    "do_recipe_sanity",
    "do_recipe_sanity_all",
    "do_rm_work",
    "do_rm_work_all",
    "do_rootfs",
    "do_rootfs_wicenv",
    "do_savedefconfig",
    "do_sdk_depends",
    "do_shared_workdir",
    "do_shared_workdir_setscene",
    "do_sizecheck",
    "do_spdx",
    "do_stash_locale",
    "do_stash_locale_setscene",
    "do_strip",
    "do_symlink_kernsrc",
    "do_testexport",
    "do_testimage",
    "do_testsdk",
    "do_testsdkext",
    "do_unpack",
    "do_unpack_and_patch",
    "do_validate_branches",
    "do_warn_musl",
    "do_write_config",
    "do_write_qemuboot_conf"
]



from oelint_adv.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarSummary80Chars(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.summary80chars",
                         severity="warning",
                         message="'SUMMARY' should not be longer than 80 characters")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="SUMMARY")
        for i in items:
            val = i.VarValueStripped
            if len(val) > 80:
                res += self.finding(i.Origin, i.InFileLine)
        return res
